# -*- coding: iso-8859-1 -*-
#
# $Id: AbstractQuestionContainer.py,v 1.28 2006/02/16 17:14:33 mxp Exp $
#
# Copyright  2004 Otto-von-Guericke-Universitt Magdeburg
#
# This file is part of LlsMultipleChoice.
#
# LlsMultipleChoice is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LlsMultipleChoice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LlsMultipleChoice; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import random

from DateTime import DateTime

from AccessControl import ClassSecurityInfo
import Acquisition
from Acquisition import *

from Products.Archetypes.public import BaseFolderSchema, Schema, \
     OrderedBaseFolder, BooleanField, IntegerField, StringField, TextField, \
     TextAreaWidget
from Products.Archetypes.Widget import TypesWidget, BooleanWidget, \
     IntegerWidget, RichWidget

from Products.CMFCore.utils import getToolByName

from Products.LlsMultipleChoice.config import PROJECTNAME, SKINS_DIR, \
     QUESTION_DIR, QUESTION_TYPES, ANSWER_DIR, \
     ANSWER_TYPES, GLOBALS, PERMISSION_INTERROGATOR, \
     PERMISSION_STUDENT, I18N_DOMAIN, CUSTOM_EVALUATION_FUNCTION_NAME
from Products.LlsMultipleChoice.CandidateTestResult import CandidateTestResult
from Products.LlsMultipleChoice.tools import getParent, filterByUID, \
     evalFunString, log, isNumeric, makeTransactionUnundoable, \
     registerTypeLogged
from Products.LlsMultipleChoice.MultipleChoiceReference \
     import MultipleChoiceReference


class ReferenceWrapper(Acquisition.Implicit):
    def __getattr__(self, key):
        #log("__getattr__(%s, %s)\n" %(str(self), str(key)))
        # Otherwise we get into an endless recursion. For whatever
        # reason ...
        if key in ["__methods__", "__members__"]:
            raise AttributeError("'%s' object has no attribute '%s'"
                                 %(repr(self), str(key)))
        else:
            return self.refObj.__of__(self.parent).aq_acquire(key)
        
    
    def __init__(self, parent, refObj):
        # don't rename the 'parent' attribute unless you change the
        # name in tools.getParent() as well!
        self.parent = parent
        self.refObj = refObj


class AbstractQuestionContainer(OrderedBaseFolder):
    """An abstract folder-like class that is designed to contain
    question objects.
        
    The derived classes are 'MultipleChoiceTest', 'QuestionGroup'.
    They contain question objects and thus have common functionality
    which they inherit from this class.
    """

    """The Archetype schema of this type.  It basically describes how
    this class looks like for Zope, i.e.  which properties 'documents'
    of this type have, which types these properties have, which
    widgets should be used for manipulation, which permissions apply
    etc.  Archetypes automatically generates accessors/mutators
    (get/set methods) for all the properties defined in the schema if
    methods with these names do not already exist.  E.g. by defining
    TextField('answer') Archetypes generates 'getAnswer()' and
    'setAnswer()'.
    """
    schema = OrderedBaseFolder.schema + Schema((
            # Read the 'description_msgid' for explanations of the
            # various properties of this type
            TextField(
                # The Python name of this property (i.e. in Python
                # code it can be accessed via 'self.directions')
                "directions",
                # This property does not have to be entered if a new
                # document of this type is created.
                required=False,
                # Text in various formats can be entered
                allowable_content_types=('text/plain',
                                         'text/structured',
                                         'text/restructured',
                                         'text/html',),
                # Output the field's contents as HTML by
                # default. Otherwise HTML elements will simply be
                # stripped.  To see formatted ouput, you also need to
                # put the keyword "structure" in front of the call to
                # the getter method, in the corresponding page
                # template. Write something like
                #
                # <span tal:replace="structure here/getDirections">
                #   Directions</span>
                # instead of <span tal:replace="here/getDirections">Directions
                #   </span>
                #
                # which is how you normally access the contents of a
                # field.
                default_output_type='text/html',
                widget=TextAreaWidget(
                    # Default label of this property.  This is the
                    # name used for this property in e.g. the
                    # base_edit view of a document of this type.
                    label='Directions',
                    # label_msgid of this property (used for
                    # internationalization)
                    label_msgid='directions_label',
                    # Default description of this property This
                    # description will be the tool tip displayed for
                    # this property in e.g. the base_edit view of a
                    # document of this type.
                    description='Some content that all the questions in '
                    'this folder refer to. You can also enter a text '
                    'which helps the candidates to better understand '
                    'the questions.',
                    # description_msgid of this property (used for
                    # internationalization)
                    description_msgid='directions_tool_tip',
                    # The internationalization domain for this
                    # property, i.e. the namespace for label_msgid and
                    # description_msgid.  This namespace is defined
                    # via "Domain: llsMultipleChoice\n" in the
                    # .po/.pot-files in the 'i18n' directory.
                    i18n_domain=I18N_DOMAIN),
                validators=('isXML',),
            ),
            BooleanField("randomOrder",
                # Specify an accessor for this property named differently
                # than the standard accessor ('getRandomOrder()' or so)
                #  that Archetypes would have generated
                accessor='isRandomOrder',
                required=False,
                # Default to 'True'
                # Working with '0' and '1' instead of 'False' and 'True'
                # seems to work better with Archetypes.
                default=1,
                # This property may not be read by user with lower
                # privileges than PERMISSION_INTERROGATOR (see
                # Products.LlsMultipleChoice.config).  That means a
                # user with PERMISSION_STUDENT will not see this
                # property when he/she # calls e.g. the base_view page
                # template.
                read_permission=PERMISSION_INTERROGATOR,
                widget=BooleanWidget(
                    label='Randomize Question Order',
                    label_msgid='randomize_question_order_label',
                    description='Check this box if you want the questions '
                    'in this container to appear in a different, random '
                    'order for each candidate. Otherwise the same order '
                    'as in the &quot;contents&quot;-view will be used.',
                    description_msgid='randomize_question_order_tool_tip',
                    i18n_domain=I18N_DOMAIN),
            ),
            IntegerField("numberOfRandomQuestions",
                required=False,
                default=-1,
                read_permission=PERMISSION_INTERROGATOR,
                widget=IntegerWidget(
                    label='Number of Random Questions',
                    label_msgid='number_of_random_questions_label',
                    description='The number of questions which are randomly '
                        'selected when a new test is '
                        'generated for a candidate. (This only works if '
                        '&quot;Randomize Question Order&quot; '
                        'is checked.) A value &lt;= 0 means that all '
                        'questions will be used.',
                    description_msgid='number_of_random_questions_tool_tip',
                    i18n_domain=I18N_DOMAIN),
            ),
        ),)

    """Register custom Page Templates for specific actions, e.g.
    'view', 'edit' or self defined actions."""
    actions = (
        {
            'id'           : 'view',
            # Use questioncontainer_view.pt instead of 'base_view' for viewing
            'action'       : 'string:${object_url}/questioncontainer_view',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
    )

    # The Zope name of this type
    meta_type = portal_type = archetype_name = "Abstract Question Container"
    """'Standalone' documents of this type are not allowed. Documents
    of this type are allowed only as part of certain other types of
    documents, i.e. those, whose 'allowed_content_types' property (see
    below) is set to something like
    'tuple(['AbstractQuestionContainer'])'.  That means you cannot
    create an 'AbstractQuestionContainer' instance directly on the
    start page but only as content in other (folder-like) documents.
    """
    global_allow = False
    """ 'AbstractQuestionContainer' documents can only contain other
    documents of the types specified in the list QUESTION_TYPES (see
    Products.LlsMultipleChoice.config).
    """
    # The following 2 statements add every class in 'QUESTION_TYPES'
    # to 'allowed_content_types'.
    #allowed_content_types = ()
    #for entry in QUESTION_TYPES:
    #    question_archetype_name = eval(
    #        "__import__(QUESTION_DIR, globals(), locals())."
    #        + entry + "." + entry + ".archetype_name" )
    #    allowed_content_types += (question_archetype_name,)
    # ... but we don't want that because 'MC Question' is the only type we 
    # really understand right now.
    allowed_content_types = ('MC Question',
                             MultipleChoiceReference.archetype_name)
        
    # Get a ClassSecurityInfo-instance in order to declare some class
    # methods protected or private
    security = ClassSecurityInfo()

    """Only users with privileges of PERMISSION_INTERROGATOR (see
    Products.LlsMultipleChoice.config) or higher may call this
    function directly or indirectly (e.g. by calling the
    folder_contents page template).
    """
    security.declareProtected(PERMISSION_INTERROGATOR, 'listFolderContents')
    """Declaring 'folderlistingFolderContents' as protected prevents
    the answers from being listed if someone without
    PERMISSION_INTERROGATOR tries to call the 'base_view' template for
    a derived type (like test or question group).
    """
    security.declareProtected(PERMISSION_INTERROGATOR,
                              'folderlistingFolderContents')


    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        OrderedBaseFolder.manage_afterAdd(self, item, container)
        # Add local role 'Manager' for the creator so that users can
        # create tests without having to be Manager
        creator = self.Creator()
        roles = list(self.get_local_roles_for_userid(creator))
        if 'Manager' not in roles:
            roles.append('Manager')
            self.manage_setLocalRoles(creator, roles)


    # The following functions would have to be implemented by derived
    # classes but through Zope's acquisition, we get exactly what we
    # want for free.

    # def getResults(self):
    # def setResults(self, results = {}):
    # def isPublic(self):
    # def getEvaluationScripts(self):
    
    
    security.declareProtected(PERMISSION_INTERROGATOR, 'getAllQuestions')
    def getAllQuestions(self):
        """Return all the Question elements (no question groups) which
        are direct children of this folder.
        """
        return filter( lambda item :
                       item.Type() in
                       AbstractQuestionContainer.allowed_content_types,
                       self.listFolderContents() )
    

    def listFolderContents(self):
        """Same as 'listFolderContents()' except that for elements of
        type 'MultipleChoiceReference', the referenced object is
        returned.
        """
        def getRefOrSelf(obj):
            if obj.Type() == MultipleChoiceReference.portal_type:
                refObj = obj.getReference()
                if refObj:
                    obj = ReferenceWrapper(self, refObj)
                else:
                    obj = None
                    
            return obj

        contents = OrderedBaseFolder.listFolderContents(self)
        ret = []
        for o in contents:
            refOrSelf = getRefOrSelf(o)
            if refOrSelf:
                ret.append(refOrSelf)
        return ret


    def _notifyOfCopyTo(self, container, op=0):
        """This method is here, because the 'reference' attribute of
        'MultipleChoiceReference' objects is not copied by default. So
        we call '_notifyOfCopyTo()' to tell every
        'MultipleChoiceReference' object to back up its 'reference'
        attribute.

        In Archetypes 1.4, this should not be necessary anymore
        because there will be switch to enable copying the contents of
        ReferenceFields.
        """
        #log("%s._notifyOfCopyTo()\n" % str(self))
        for obj in self.contentValues():
            obj._notifyOfCopyTo(self, op=op)
        return OrderedBaseFolder._notifyOfCopyTo(self, container, op=op)


    setResultSecure = False
    def setResult(self, candidateId, candidateTestResult):
        """Adds the candidateTestResult to the results dictionary of
        the 'MultipleChoiceTest' or deletes the result for
        'candidateId' if 'candidateTestResult' is 'None'.  This method
        is needed because changing the 'results' property of the
        MultipleChoiceTest directly does not work for some
        reason. Changes only work if the entire dictionary is written
        back.

        @param candidateId The user ID.
            
        @param candidateTestResult The test result of the candidate
        with user ID candidateId.  If 'None', the former result for
        'candidateId' is deleted.
        """
        while(self.setResultSecure): # FIXME: Race condidition!
            pass
        self.setResultSecure = True
        try:
            results = self.getResults()
            if(candidateTestResult):
                results[candidateId] = candidateTestResult
            elif( (candidateTestResult is None) \
                and results.has_key(candidateId) ):
                results.pop(candidateId)
            # Save the chages to the results in the
            # 'MultipleChoiceTest' instance
            self.setResults(results)
        finally:
            self.setResultSecure = False


    def setCandidateAnswer(self, candidateId, questionId, answerValue):
        """Save the answer of a candidate to a question in the
        'results' property of the 'MultipleChoiceTest' instance.  This
        method is called by the 'setCandidateAnswer()' of the question
        object which is in turn called by the 'submitTest' script in
        the skins directory.

        @param candidateId The user ID of the candidate who gave the
        answer.
            
        @param questionId The ID (short name) of the question that is
        answered.
            
        @param answerValue The actual answer. This will be whatever
        'request.get(questionId)' returned, i.e. the value or the list
        of values of the <input>-element(s) with the name questionId.
        """
        #log('AbstractQuestionContainer.setCandidateAnswer(%s, %s, %s)\n' %(str(candidateId), str(questionId), str(answerValue)))
        if(not candidateId):
            # Do not save submissions from users who are not logged in
            return

        results = self.getResults()
        if(results.has_key(candidateId)):
            candidateTestResult = results[candidateId]
        else:
            # no such candidate? --> create new CandidateTestResult instance
            candidateTestResult = CandidateTestResult()
        candidateTestResult.setCandidateAnswer(questionId, answerValue)
        # Save the chages to the results in the 'MultipleChoiceTest' instance
        self.setResult(candidateId, candidateTestResult)


    def getCandidateAnswer(self, candidateId, questionId):
        """Returns the answer that a candidate gave (selected) to
        (for) a question.  The return value will be whatever
        'setCandidateAnswer()' (s.a.) saved for that question. For the
        standard answer type this can be the ID of the selected
        answer, a list of IDs of selected answers or 'None', e.g.  if
        there is no candidate with user ID candidateId.

        @param candidateId The user ID of the candidate who gave the
        answer.
            
        @param questionId The ID (short name) of the question that is
        answered.
        """
        #log('AbstractQuestionContainer.getCandidateAnswer(%s, %s)\n' %(candidateId, questionId))
        if(not candidateId):
            return None
        results = self.getResults()
        if(not results.has_key(candidateId)):
            #log('\treturn None\n')
            return None
        #elif(not (questionId in results[candidateId].getQuestionIds())):
            #return None
        else:
            #log('\treturn ' + str(results[candidateId].getCandidateAnswer(questionId)) + '\n')
            return results[candidateId].getCandidateAnswer(questionId)


    def getQuestions(self, candidateId):
        """Returns the actual question objects that the candidate saw
        in his/her test and in the order they were presented
        """
        #log('AbstractQuestionContainer.getQuestions(%s)\n' %str(candidateId))
        if(not candidateId):
            return []
        results = self.getResults()
        allQuestions = self.getAllQuestions()
        #log('\tallQuestions = ' + str(allQuestions) + '\n')
        if(not self.isPublic()):
            questions = allQuestions
        else:
            if( (not results.has_key(candidateId))
                or (not (self.getId() in
                         results[candidateId].getQuestionContainerIds())) ):
                # generating a new test
                #log('\tgenerating a new test\n')
                if(self.isRandomOrder()):
                    # use random order
                    if(self.getNumberOfRandomQuestions() > 0):
                        randomQuestions = random.sample(
                            allQuestions,
                            min( len(allQuestions),
                                 self.getNumberOfRandomQuestions() ) )
                        questions = randomQuestions
                    else:
                        questions = random.sample(allQuestions,
                                                  len(allQuestions))
                else:
                    # use the order in the "contents"-view
                    questions = allQuestions
            else:
                # reconstructing the old test
                #log('\treconstructing the old test\n')
                questionIds = results[candidateId].getQuestionIds()
                # reconstruct the order of the questions
                questions = filterByUID(questionIds, allQuestions)
        #log('\t   questions = ' + str(questions) + '\n')
        return questions


    def getAnswers(self, candidateId, questionId):
        """Returns the IDs (or whatever was saved) of the answers that
        the candidate saw for the question with ID questionId.

        @param candidateId The user ID of the candidate.
            
        @param questionId The ID (short name) of the question.
        """
        # log('AbstractQuestionContainer.getAnswers(%s, %s)\n'
        #     %(str(candidateId), str(questionId)))
        if(not candidateId):
            return []
        results = self.getResults()
        if(not results.has_key(candidateId)):
            return None
        elif(not questionId in results[candidateId].getQuestionIds()):
            return None
        else:
            return results[candidateId].getAnswers(questionId)


    def setAnswers(self, candidateId, questionId, suggestedAnswerValues):
        """Saves the IDs of the answers that are suggested to the
        candidate for the question with ID questionId.

        @param candidateId The user ID of the candidate.
            
        @param questionId The ID (short name) of the question.
            
        @param suggestedAnswerValues A list of IDs of the answers that
        were suggested to the candidate."""
        #log('AbstractQuestionContainer.setQuestions(%s, %s, %s)\n'
        #    %(str(candidateId), str(questionId), str(suggestedAnswerValues)))
        if((not candidateId) or (not self.isPublic())):
            return

        results = self.getResults()
        if(results.has_key(candidateId)):
            # get the CandidateTestResult instance for the candidate
            candidateTestResult = results[candidateId]
        else:
            # no such candidate? --> create new CandidateTestResult instance
            candidateTestResult = CandidateTestResult()
            candidateTestResult.setTimeStart(DateTime())
        # Save suggestedAnswerValues
        candidateTestResult.setAnswers(self.getId(), questionId,
                                       suggestedAnswerValues)
        # Save the changes to the results in the 'MultipleChoiceTest' instance
        self.setResult(candidateId, candidateTestResult)
        # Make this un-undoable by the candidate
        makeTransactionUnundoable()


    def getPossiblePoints(self, candidateId):
        """Return how many points the candidate could have got, which
        is the sum of points for all the questions in this container
        (not recursively).

        @param candidateId the user ID of the candidate whose points
        you want to know.
        """
        questions = self.getQuestions(candidateId)
        points = 0
        for question in questions:
            if(not hasattr(question, 'getPoints')):
                return None
            questionPoints = question.getPoints()
            if(not isNumeric(questionPoints)):
                return None
            points = points + questionPoints
        return points


    def getEvaluationScript(self, archetype_name):
        """Return the custom evaluation script that has been uploaded
        for the class with Archetype Name archetype_name or None if no
        script has been uploaded.

        @param archetype_name The Archetype Name of the class.
        """
        return None

        
        evaluationScripts = self.getEvaluationScripts()
        if(evaluationScripts.has_key(archetype_name)):
            return evaluationScripts[archetype_name]
        else:
            return None

    def getEvaluationScriptComment(self, archetype_name):
        """TODO"""
        script = self.getEvaluationScript(archetype_name)
        if script:
            d = {}
            code = compile(script, '<string>', 'exec')
            exec code in d
            return d[CUSTOM_EVALUATION_FUNCTION_NAME].__doc__
        else:
            return None


# Register this type in Zope
registerTypeLogged(AbstractQuestionContainer)
