# -*- coding: iso-8859-1 -*-
#
# $Id: CandidateTestResult.py,v 1.12 2006/01/31 18:17:21 wfenske Exp $
#
# Copyright  2004 Otto-von-Guericke-Universitt Magdeburg
#
# This file is part of LlsMultipleChoice.
#
# LlsMultipleChoice is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LlsMultipleChoice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LlsMultipleChoice; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import datetime
from Products.LlsMultipleChoice.tools import log

class CandidateTestResult:
    """ This class contains all the candidate-specific information for
        a multiple choice test. They are:
        
        answers:

          A dictionary with all the answers that were suggested to the
          candidate for a question. The keys are the IDs of the
          questions. With the standard question and answer types the
          value for a given key is the ID or a list of the IDs of the
          suggested answer(s). Custom question/answer types can save
          anything the want.
                                   
        questionContainerIds:

          A list of the IDs of the 'AbstractQuestionContainer'
          instances that the candidate saw.

        questionIds:

          A list of IDs of the questions that the candidate saw in
          his/her test. Unfortunately 'answers.keys()' cannot be used
          since it does not preserve the order of the IDs.
                                   
        candidateAnswers:

          The dictionary containing all the answers which the
          candidate thought were correct (usually the ones which
          he/she selected). The keys are the IDs of the
          questions. With the standard question and answer types the
          value for a given key is the ID or a list of the IDs of the
          selected answer(s). Custom question/answer types can save
          anything the want.
                                   
        timeStart:

          The time when the candidate first saw the test.
        
        timeFinish:

          The time when the candidate submitted his/her test.
    """    
    
    def __init__(self):
        """ Initializes all the properties with empty values. """
        self.answers = {} # explanation s.a.
        self.questionContainerIds = [] # explanation s.a.
        self.questionIds = [] # explanation s.a.
        self.candidateAnswers = {} # explanation s.a.
        self.timeStart = None # explanation s.a.
        self.timeFinish = None # explanation s.a.
        
        
    def getTimeStart(self):
        return self.timeStart
        
    def setTimeStart(self, dateTimeObject):
        self.timeStart = dateTimeObject
        
    def getTimeFinish(self):
        return self.timeFinish
        
    def setTimeFinish(self, dateTimeObject):
        self.timeFinish = dateTimeObject
    
    
    def getQuestionContainerIds(self):
        """ Returns the IDs of all the question containers, questions
            of which the candidate was asked.
        """
        return self.questionContainerIds
        
    
    def getQuestionIds(self):
        """ Returns the IDs of all the questions the candidate was asked. """
        return self.questionIds
    
    
    def setAnswers(self, questionContainerId, questionId,
                   suggestedAnswerValues):
        """ Saves the suggested answer values to a question the
            candidate was asked.
            
            @param suggestedAnswerValues The value or list of values
            of the answers suggested to the candidate.
                                        
            @param questionId The ID of the question that the answers
            belong to.
                              
            @param questionContainerId The ID of the question
            container that contains the question.
        """
        # The list 'questionContainerIds' is neccessary because 
        # 'self.answers.keys()' does not return the questionIds in the order 
        # they were added.
        if(not questionContainerId in self.getQuestionContainerIds()):
            self.questionContainerIds.append(questionContainerId)
        if(not questionId in self.getQuestionIds()):
            self.questionIds.append(questionId)
        self.answers[questionId] = suggestedAnswerValues
    
    
    def getAnswers(self, questionId):
        """ Returns the suggested answer values for a question the
            candidate was asked.  With the standard question and
            answer types this will be a list of IDs of answer objects.
            
            @param questionId The ID of the question.
        """
        if(questionId in self.getQuestionIds()):
            return self.answers[questionId]
        else:
            txt = "getAnswers() called for a questionId which is not " \
                  "in in self.getQuestionIds()\n"
            log(txt)
            raise Exception(txt)
        
    def setCandidateAnswer(self, questionId, answerValue):
        """ Saves the candidate's answer to a question.
            
            @param answerValue The value or list of values of the
            answer(s) the candidate gave. With the standard question
            and answer types this will be the ID (list of IDs) of the
            answer(s) the candidate selected.
                              
            @param questionId The ID of the question that the answer
            belongs to.
        """
        if(not questionId in self.getQuestionIds()):
            txt = "setCandidateAnswer() called for questionId not in " \
                  "self.getQuestionIds()\n"
            log(txt)
            raise Exception(txt)
        else:
            self.candidateAnswers[questionId] = answerValue
        
    def getCandidateAnswer(self, questionId):
        """ Returns the candidate's answer to the question with ID
            questionId.  If this question was not part of his/her
            test, an exception will be raised.
            
            @param questionId The ID of the question.
        """
        if(not questionId in self.getQuestionIds()):
            txt = "getCandidateAnswer() called for questionId not in " \
                  "self.getQuestionIds()\n" \
                  "\tquestionId is " + str(questionId) + "\n"
            log(txt)
            #~ raise Exception(txt)
            return None
        else:
            if(self.candidateAnswers.has_key(questionId)):
                return self.candidateAnswers[questionId]
            else:
                return None
            
