# -*- coding: iso-8859-1 -*-
#
# $Id: MultipleChoiceTest.py,v 1.38 2006/02/28 16:31:06 mxp Exp $
#
# Copyright  2004 Otto-von-Guericke-Universitt Magdeburg
#
# This file is part of LlsMultipleChoice.
#
# LlsMultipleChoice is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LlsMultipleChoice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LlsMultipleChoice; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, os.path
from AccessControl import ClassSecurityInfo, getSecurityManager
from Acquisition import *
from DateTime import DateTime
from zipfile import ZipFile
from Products.Archetypes.public import Schema, BooleanField, IntegerField, \
     ObjectField, LinesField, FileField, StringField
from Products.Archetypes.Widget import TypesWidget, BooleanWidget, \
     SelectionWidget
from Products.Archetypes.utils import DisplayList
from Products.CMFCore.utils import getToolByName
try: # Plone 2.0.5
    from Products.CMFPlone.PloneUtilities import localized_time
except: # Plone 2.1
    from Products.CMFPlone.utils import ulocalized_time as localized_time
from Products.LlsMultipleChoice.config import *
from Products.LlsMultipleChoice.AbstractQuestionContainer \
     import AbstractQuestionContainer
from Products.LlsMultipleChoice.QuestionTypes.MultipleChoiceQuestion \
     import MultipleChoiceQuestion
from Products.LlsMultipleChoice.QuestionGroup import QuestionGroup
from Products.LlsMultipleChoice.MultipleChoiceReference \
     import MultipleChoiceReference
from Products.LlsMultipleChoice.tools import getParent, evalFunString, \
     log, isNumeric, MyStringIO, makeTransactionUnundoable, registerTypeLogged
from Products.LlsMultipleChoice.qti import importAssessmentItem, \
     importPackage, exportPackage

from Statistics import Statistics


class ResultsWidget(TypesWidget):
    """ A custom widget for handling a dictionary of 
        CandidateTestResult objects. 
    """
    _properties = TypesWidget._properties.copy()
    _properties.update({
        # The Page Template for this widget
        'macro' : 'results_widget',
    })
    
    
class EvaluationScriptsWidget(TypesWidget):
    """ A custom widget for handling the 'evaluationScripts' 
        property of the MultipleChoiceTest.
    """
    _properties = TypesWidget._properties.copy()
    _properties.update({
        'macro' : 'evaluation_scripts_widget',
    })


class MultipleChoiceTest(AbstractQuestionContainer):
    """A multiple-choice test."""
    
    """This is the main class of the 'LlsMultipleChoice' Product.
    A multiple choice test is basically a folder that contains 
    questions and groups of questions, i.e. folders that
    contain questions."""
    
    # format: [file extension, column delimiter, row delimiter, 
    # start of text delimiter, end of text delimiter, escape char]
    RESULTS_EXPORT_FORMAT_TAB = ['tab', '\t', '\n', '"', '"', '"']
    RESULTS_EXPORT_FORMAT_CSV = ['csv', ';',  '\n', '"', '"', '"']
    RESULTS_EXPORT_FORMATS = [RESULTS_EXPORT_FORMAT_TAB, RESULTS_EXPORT_FORMAT_CSV]
    
    # See comments in 'AbstractQuestionContainer' for an explanation
    # of the function of a schema, therein defined properties (fields) and 
    # internationalization of the widgets
    schema = AbstractQuestionContainer.schema + Schema((
            ObjectField('results',
                        # A dictionary that contains CandidateTestResult
                        # objects.  The keys are the candidate IDs. The
                        # value for a key is the corresponding
                        # CandidateTestResult instance. (That means that
                        # you cannot save more than one
                        # CandidateTestResult instance for a candidate.
                        required=False,
                        # Not sure what this does.
                        multiValued=True,
                        default={},
                        mutator='setResults',
                        widget=ResultsWidget(),
                        read_permission=PERMISSION_INTERROGATOR,
            ),
            BooleanField('instantFeedback',
                        # Instant feedback means that the candidate will see
                        # their results immediately after submitting their
                        # test.
                        required=False,
                        default=0,
                        accessor='isInstantFeedback',
                        widget=BooleanWidget(
                            label='Instant Feedback',
                            label_msgid='instant_feedback_label',
                            description='If you want to give the candidates '
                            'instant feedback check this box.',
                            description_msgid='instant_feedback_tool_tip',
                            i18n_domain=I18N_DOMAIN),
            ),
            BooleanField('allowRepetition',
                        # See 'description' of the widget.
                        required=False,
                        default=0,
                        accessor='isAllowRepetition',
                        widget=BooleanWidget(
                            label='Allow Repetition',
                            label_msgid='allow_repetition_label',
                            description='If you want to allow repeated '
                            'submission of the test check this box.',
                            description_msgid='allow_repetition_tool_tip',
                            i18n_domain=I18N_DOMAIN),
            ),
#             ObjectField('evaluationScripts',
#                         # A dictionary for custom evaluation
#                         # scripts. It is possible to customize the
#                         # evaluation behavior of the LlsMultipleChoice
#                         # Product by uploading custom evaluation
#                         # scripts for 1) the MultipleChoiceTest, 2)
#                         # QuestionGroup objects, 3) Question
#                         # objects. Otherwise a default evaluation
#                         # method will be used.  The keys of this
#                         # dictionary are the'archetype_name' properties of 
#                         # those classes. The values will be the scripts.
#                         required=False,
#                         # Not sure what this does.
#                         multiValued=True,
#                         default={},
#                         mutator='setEvaluationScripts',
#                         widget=EvaluationScriptsWidget(
#                             types=CUSTOM_EVALUATION_TYPES),
#                         read_permission=PERMISSION_INTERROGATOR,
#             ),
            StringField('scoringFunction',
                        default='guessing',
                        enforceVocabulary=1,
                        vocabulary=DisplayList((
                            ('guessing', 'Guessing Correction',
                             'scoring_fun_guessing_correction_label'),
                            ('cruel', 'All or Nothing',
                             'scoring_fun_cruel_label'),
                            )),
                        widget=SelectionWidget(
                            label='Scoring Function',
                            label_msgid='scoring_fun_label',
                            description='The way the score for a question '
                            'is calculated.',
                            description_msgid='scoring_fun_tool_tip',
                            i18n_domain=I18N_DOMAIN),
            ),),)

    """ Inherit custom actions from 'AbstractQuestionContainer' and redefine
        them and/or define some new ones. 
    """
    actions = AbstractQuestionContainer.actions + (
        {
            'id'           : 'view',
            'action'       : 'string:${object_url}/test_view',
            'permissions'  : (PERMISSION_STUDENT,),
        },
#         {
#             'id'           : 'upload_evaluation_scripts',
#             'name'         : 'Script Upload',
#             'action'       : 'string:${object_url}/upload_evaluation_scripts',
#             'permissions'  : (PERMISSION_INTERROGATOR,),
#         },
        {
            'id'           : 'test_import_export',
            'name'         : 'Import/Export',
            'action'       : 'string:${object_url}/test_import_export',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
        {
            'id'           : 'see_results',
            'name'         : 'Results',
            'action'       : 'string:${object_url}/test_results',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
        {
            'id'           : 'statistics',
            'name'         : 'Statistics',
            'action'       : 'string:${object_url}/test_statistics',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
    )
    
    meta_type = portal_type = archetype_name = 'MC Test'
    # This type is directly allowed anywhere.
    global_allow = True
    # MultipleChoiceTests may only contain questions and groups of question.
    allowed_content_types = (QuestionGroup.archetype_name, 'Folder', 'File',
                             'Image') \
                             + AbstractQuestionContainer.allowed_content_types
    
    """ A custom icon for MultipleChoiceTests. The icon is located in 
        the skins directory ('skins/llsMultipleChoice') (It is possible to put
        it in another folder but the path must be known to Zope.). 
    """
    content_icon = 'MultipleChoiceTest.png'
    
    security = ClassSecurityInfo()
    
    
    def __getattr__(self, key):
        """ This workaround method maps  read requests for the standard 
            Plone properties 'effective_date'  and 'expiration_date' to
            calls to  the Archetype  methods  'getEffectiveDate()'  and 
            'getExpirationDate()',  respectively. This way timed publi-
            shing/unpublishing works with MultipleChoiceTests, too.
              Consult  the Python  documentation  to  learn  more about
            '__getattr__()' methods.
        """
        if(key == 'effective_date'):
            return self.getEffectiveDate()
        elif(key == 'expiration_date'):
            return self.getExpirationDate()
        else:
            raise AttributeError("'%s' object has no attribute '%s'"
                                 %(repr(self), str(key)) )



    
    security.declarePrivate('getReferencedObjects')
    def getReferencedObjects(self):
        """Collect all objects referenced by a MultipleChoiceReference
        object within this test (and its QuestionGroups).
        """
        tmp = self.contentValues()
        all = list(tmp)
        for o in tmp:
            if o.Type() == QuestionGroup.portal_type:
                all.extend(o.contentValues())
        return [o.getReference() for o in all
                if o.Type() == MultipleChoiceReference.portal_type]
    

    security.declarePrivate('getTests')
    def getTests(self, referencedObjects):
        """Returns a list of the original MultipleChoiceTests that the
        objects in 'referencedObjects' come from.
        """
        ret = []
        for o in referencedObjects:
            while True:                            
                o = aq_parent(o)
                if o:
                    if o.Type() \
                           == MultipleChoiceTest.portal_type:
                        if not o in ret:
                            ret.append(o)
                        break
                else:
                    break
        return ret


    def __bobo_traverse__(self, REQUEST, name):
        """Through this function, images, files, etc. that were
        referenced by MultipleChoiceReference objects in the test get
        found.

        It works by searching for the files in this test ('self') and
        in every test from which some question or question group was
        referenced.
        """
        try:
            if REQUEST.has_key('ACTUAL_URL'):
                referencedObjects = self.getReferencedObjects()
                tests = self.getTests(referencedObjects)
                #log("TESTS: %s\n" % str(tests))
                
                # we only need to do something if we got any new tests
                if tests:
                    # first, search in 'self', then try the others
                    tests = [self] + tests
                    # collect the values that need replacing
                    repl = {}
                    self_url = self.absolute_url()
                    for key in REQUEST.keys():
                        value = REQUEST[key]
                        if (type(value) in [str, unicode]) and \
                               (value.find(self_url) != -1):
                            repl[key] = value
                    
                    # try to get the requested object from one of the
                    # tests in 'tests'
                    for test in tests:
                        # rewrite the URL
                        for key in repl.keys():
                            REQUEST[key] = repl[key].replace(
                                self_url, test.absolute_url())
                        # try to get the object using the rewritten URL
                        try:
                            result = \
                                AbstractQuestionContainer.__bobo_traverse__(
                                    test, REQUEST, name)
                        except Exception, e:
                            #log("Exception: %s\n" % str(e))
                            result = None
                        # return the object if we got anything
                        if result:
                            return result
        except Exception, e:
            #log("Exception: %s\n" % str(e))
            pass
            
        return AbstractQuestionContainer.__bobo_traverse__(self, REQUEST, name)
    
    
    def hasParticipated(self, candidateId):
        """ Find out if a test has been generated for candidate "candidateId", 
            if he has broken the seal, so to speak. 
        """
        return self.getResults().has_key(candidateId)
        
    
    #~ security.declareProtected(PERMISSION_INTERROGATOR, 'getParticipantIds')
    def getParticipantIds(self):
        """ Get the IDs of all the candidates for whom a test has been
            generated.
        """
        return self.getResults().keys()
        
        
    def hasSubmitted(self, candidateId):
        """ Find out if a candidate has actually submitted/taken this test. """
        #~ log("MultipleChoiceTest.hasSubmitted():\n")
        if(self.results.has_key(candidateId)):
            return (self.getCandidateTimeFinish(candidateId) != None)
        else:
            #~ log("\tnot self.results.has_key(" + candidateId + ")\n")
            return False
            
    
    def getCandidateTimeStart(self, candidateId):
        """ Returns the time when the candidate opened his/her test.
            (The start time is set in AbstractQuestionContainer.setAnswers())
        """
        #~ log('MultipleChoiceTest.getCandidateTimeStart()\n')
        if(self.results.has_key(candidateId)):            
            timeStart = self.results[candidateId].getTimeStart()
            return timeStart
        else:
            errorText = 'MultipleChoiceTest.getCandidateTimeStart()\nError: '\
                'no results for candidate with ID ' + candidateId
            log(errorText)
            return None
            
            
    def setCandidateTimeFinish(self, candidateId, dateTimeObject):
        """ Save the time when the candidate submitted his/her test. """
        #~ log('MultipleChoiceTest.setCandidateTimeFinish()\n')
        if(self.results.has_key(candidateId)):
            self.results[candidateId].setTimeFinish(dateTimeObject)
        else:
            errorText = 'MultipleChoiceTest.setCandidateTimeFinish()\nError: '\
                'no results for candidate with ID ' + candidateId
            log(errorText)
        # Make this un-undoable by the candidate
        makeTransactionUnundoable()
            
            
    def getCandidateTimeFinish(self, candidateId):
        """ Returns the time when the candidate submitted his/her test. """
        #~ log('MultipleChoiceTest.getCandidateTimeFinish()\n')
        if(self.results.has_key(candidateId)):
            timeFinish = self.results[candidateId].getTimeFinish()
            return timeFinish
        else:
            errorText = 'MultipleChoiceTest.getCandidateTimeFinish()\nError: '\
                'no results for candidate with ID ' + candidateId
            log(errorText)
            return None
    
    #~ security.declareProtected(PERMISSION_INTERROGATOR, 'getSubmitterIds')
    def getSubmitterIds(self):
        """ Return the IDs of the candidates who have actually submitted/taken
            this test.
        """
        return filter((lambda candidateId : self.hasSubmitted(candidateId)),
                      self.results.keys())
    
    #~ security.declareProtected(PERMISSION_INTERROGATOR, 'getResults')
    security.declarePrivate('setResults')
    # This function has to be implemented by classes derived from 
    # AbstractQuestionContainer
    def setResults(self, results = {}):
        """ Save the dictionary of CandidateTestResult objects in this test 
            instance. 
        """
        #~ log('MultipleChoiceTest.setResults()\n')
        if( type(results) != dict ):
            results = {}
        self.results = results
        #~ log('\tself.results.keys() = %s\n' % str(self.results.keys()))
        
    
    security.declareProtected(PERMISSION_INTERROGATOR, 'setEvaluationScripts')
    def setEvaluationScripts(self, evaluationScripts = {}):
        """ Save the dictionary of custom evaluation scripts in this 
            instance. 
        """
        if( type(evaluationScripts) == dict ):
            self.evaluationScripts = evaluationScripts
        else:
            evaluationScripts = {}
            
        
    security.declareProtected(PERMISSION_INTERROGATOR,
                              'processEvaluationScriptUpload')
    def processEvaluationScriptUpload(self, archetype_name, funString):
        
        """ Routine for uploading custom evaluation scripts.
        
            The scripts are stored in the ObjectField
            'evaluationScripts' of the test instance. (That implies
            every test instance can have its own evaluation script.)
            
            @param archetype_name The archetype name of the class you
                   want to customize, e.g. 'Question' or 'MC Test'.
                   
            @param funString A string containing a python script that
                   can be compiled with Python's 'compile'
                   function. The script has to define an evaluation
                   function.  The name of that function has to be what
                   is specified in the global constant
                   CUSTOM_EVALUATION_FUNCTION_NAME (see config.py).
                   See the last line of the 'getCandidatePoints()'
                   methods of 'MultipleChoiceTest', 'QuestionGroup'
                   and 'Question' for the parameters these functions
                   must accept.
        """
        
        #~ log("MultipleChoiceTest.processEvaluationScriptUpload(\n%s, \n%s)\n"
        #~     %(str(archetype_name), funString))
        try: # Try to compile the script
            codeObj = compile(funString, '<string>', 'exec')
        except Exception, e:
            # Can't be compiled --> not useable
            errorText = str(e)
            return errorText
        noFunctionDefinition = True
        # Check if the script defines a function whose name is equal to the 
        # value of the constant 'CUSTOM_EVALUATION_FUNCTION_NAME'
        if CUSTOM_EVALUATION_FUNCTION_NAME in codeObj.co_names:
            exec(codeObj)
            # Check if it is really a function and not a class or var
            if getattr(eval(CUSTOM_EVALUATION_FUNCTION_NAME),
                       'func_name', None) == CUSTOM_EVALUATION_FUNCTION_NAME:
                noFunctionDefinition = False
        if noFunctionDefinition:
            # Either nothing called 'CUSTOM_EVALUATION_FUNCTION_NAME' 
            # is defined or it's not a function
            return self.translate(
                msgid   = 'uploadEvaluationScript_no_function_definition',
                domain  = I18N_DOMAIN,
                default = 'The script does not define a function called "%s"')\
                % CUSTOM_EVALUATION_FUNCTION_NAME
        # Save the script in the test's 'evaluationScripts' dictionary
        evaluationScripts = self.getEvaluationScripts()
        evaluationScripts[archetype_name] = funString
        self.setEvaluationScripts(evaluationScripts)
##        log("self.evaluationScripts = %s\n"
##            %(str(self.getEvaluationScripts())))
        return None # return no error message (Success)
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'processQTIImport')
    def processQTIImport(self, file):
        """ Routine for uploading tests. 
        
            @param file A file containing a QTI assessmentItem or
                    a QTI package in a zip file.
        """
        addedObjects = []
        errorString = ''
        errors = MyStringIO()
        # try zipfile first
        zipFileInstance = None
        try:
            zipFileInstance = ZipFile(file.filename)
        except:
            try:
                zipFileInstance = ZipFile(file)
            except:
                pass
        if(zipFileInstance):
            # is a zip file
            addedObjects = importPackage(self, zipFileInstance, errors)
        else:
            # no zipfile --> maybe XML?
            # reset the file read ptr first
            try:
                file.seek(0)
            except:
                try:
                    file.filename.seek(0)
                except:
                    pass
            string = None
            try:
                string = file.filename.read()
            except:
                try:
                    string = file.read()
                except:
                    pass
            if(not (type(string) in [str, unicode])):
                # give up, no XML either
                # The file could not be read.
                errors.write( '\n' + self.translate(\
                    msgid   = 'file_read_error',\
                    domain  = I18N_DOMAIN,\
                    default = 'The file could not be read.') )
            else:
                # is an XML file
                addedObjects = importAssessmentItem(self, string, errors)
        # Reset the read ptr of the StringIO obj
        errors.seek(0)
        errorString = errors.read().strip()
        #~ log(errorString + '\n')
        return addedObjects, ([errorString, None][not errorString])
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'processQTIImport')
    def processQTIExport(self):
        """ Routine for downloading tests.
        """
        errors = MyStringIO()
        package = exportPackage(self, errors)
        # Reset the read ptr of the StringIO obj
        errors.seek(0)
        errorString = errors.read().strip()
        #~ log(errorString + '\n')
        return package, ([errorString, None][not errorString])
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'getResultsAsList')
    def getResultsAsList(self, candidateIdList = None):
        """ Routine for exporting candidate results.
            
            @param candidateIdList A list of IDs of the candidates            
                whose results you want to know. If 'candidateIdList'
                is 'None', all results will be returned.
            
            @return A list of containing the results as a list of
                values. The values appear in the following order:
                
                [CandidateID, Candidate_Name, Score, Max. Score, Time
                 Started, Time Finished]
                
                The first"result" ('getResultsAsList()[0]') is a header row 
                containing strings such as 'Candidate', 'Name', 'Score' etc.. 
                The actual results start at 'getResultsAsList()[1]'.
        """
        #~ log("MultipleChoiceTest.getResultsAsList()\n")
        """
            <th i18n:translate="candidate" 
                tal:content="python:headerRow[1]">Candidate</th>
            <th i18n:translate="test_results_score" 
                tal:content="python:headerRow[2]">Score</th>
            <th i18n:translate="test_results_max_score" 
                tal:content="python:headerRow[3]">Max. Score</th>
            <th i18n:translate="started" 
                tal:content="python:headerRow[4]">Started</th>
            <th i18n:translate="finished" 
                tal:content="python:headerRow[5]">Finished</th>
        """
        llsmc_tool = getToolByName(self, 'llsmc_tool')

        # The header row
        retVal = [[
            self.translate(msgid = 'candidate', 
                      domain  = I18N_DOMAIN, 
                      default = 'Candidate'), 
            self.translate(msgid = 'candidate', 
                      domain  = I18N_DOMAIN, 
                      default = 'Candidate'), 
            self.translate(msgid = 'test_results_score', 
                      domain  = I18N_DOMAIN, 
                      default = 'Score'), 
            self.translate(msgid = 'test_results_max_score', 
                      domain  = I18N_DOMAIN, 
                      default = 'Max. Score'), 
            self.translate(msgid = 'started', 
                      domain  = I18N_DOMAIN, 
                      default = 'Started'), 
            self.translate(msgid = 'finished', 
                      domain  = I18N_DOMAIN, 
                      default = 'Finished')]]
        
        participants  = self.getParticipantIds()
        if candidateIdList is not None:
            # If not all results are requested, filter 'candidateIdList' so 
            # that every 'candidateId' in 'candidateIdList' is actually a 
            # participant, i.e. is in 'self.getParticipantIds()'
            participants = filter((lambda candidateId :
                (candidateId in participants)), candidateIdList)
        # We only want submitters who are also in the list of participants 
        # (i.e. in 'participants')
        submitters    = filter((lambda candidateId :
            (candidateId in participants)), self.getSubmitterIds())
        # Participants who have not submitted their tests
        nonSubmitters = filter((lambda candidateId :
            not (candidateId in submitters)), participants)
        
        # Sort 'participants' so that 'submitters' come first
        submitters.sort(lambda a, b: cmp(llsmc_tool.getFullNameById(a),
                                         llsmc_tool.getFullNameById(b)))
        nonSubmitters.sort(lambda a, b: cmp(llsmc_tool.getFullNameById(a),
                                         llsmc_tool.getFullNameById(b)))

        participants  = submitters + nonSubmitters
        results       = self.getResults()
        NaN           = '-' # Not a number/not available
        
        def formatTime(dateTimeObj):
            # 'localized_time' is a Plone function
            return localized_time( dateTimeObj, long_format = True,
                                   context = self )
        
        # The content rows
        for participantId in participants:
            points = None
            if(participantId in submitters):
                points       = self.getCandidatePoints(participantId)
                timeFinished = formatTime( self.getCandidateTimeFinish(
                                           participantId) )
            else:
                timeFinished = self.translate( msgid   = 'not_submitted',
                                               domain  = I18N_DOMAIN,
                                               default = 'Not submitted' )
            possiblePoints = self.getPossiblePoints(participantId)
            if(points is None):
                points = NaN
            if(possiblePoints is None):
                possiblePoints = NaN
            
            retVal.append( [
                participantId, 
                llsmc_tool.getFullNameById(participantId), 
                points, 
                possiblePoints, 
                formatTime( self.getCandidateTimeStart(participantId) ), 
                timeFinished] )

        return retVal
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'getItemStatisticsW')
    def getItemStatisticsW(self, candidateIdList = None):
        """ Routine for exporting detailed candidate results.
            
            @return A list of containing the results as a list of values. The
                values appear in the following order:
                
                [CandidateID, Candidate_Name] followed by either -, 0 or 1 for
                each answer to each question in each question group in the
                test.
                
                "-" means the candidate did not have this answer presented to
                him/her as a possibility, "0 "means he/she did see this answer
                but did not check it and "1" means he/she did check it.
                
                The first three "results" are header rows.
        """
        llsmc_tool = getToolByName(self, 'llsmc_tool')

        # The header row
        retVal = []
        h1 = ['', '']
        h2 = ['', '']
        h3 = [self.translate(msgid = 'candidate',
                             domain  = I18N_DOMAIN,
                             default = 'Candidate'),
              self.translate(msgid = 'name',
                             domain  = I18N_DOMAIN,
                             default = 'Name')]
        for group in [self] + self.getQuestionGroups():
            h1.append([group.title_or_id(),
                       self.translate(msgid = 'ungrouped',
                                      domain  = I18N_DOMAIN,
                                      default = '(Ungrouped)')
                       ][group is self])
            allQuestions = group.getAllQuestions()
            numQuestions = len(allQuestions)
            for question in allQuestions:
                numAnswers = len(question.listFolderContents())
                h1 += ['' for i in range(numAnswers)]
                h2 += [question.title_or_id()] \
                     + ['' for i in range(numAnswers-1)]
                h3 += ['a%d' % (i+1) for i in range(numAnswers)]
            # we added one placeholder too much in the loop above
            h1 = h1[:-1]
        
        for h in [h1, h2, h3]:
            retVal.append( h )
            
        # The content rows
        submitters = self.getSubmitterIds()
        if candidateIdList is not None:
            # If not all results are requested, filter 'submitters' so that
            # every 'candidateId' in 'submitters' is also in 'candidateIdList'
            submitters = filter((lambda candidateId :
                                 (candidateId in candidateIdList)), submitters)
        submitters.sort()
        NaN           = '-' # Not a number/not available
        for submitterId in submitters:
            row = [submitterId, llsmc_tool.getFullNameById(submitterId)]
            for group in [self] + self.getQuestionGroups():
                for question in group.getAllQuestions():
                    qUid = question.UID()
                    sSuggestedAnswerIds = group.getAnswers(
                        submitterId, qUid) or []
                    #log(str(sSuggestedAnswerIds))
                    sCandidateAnswerIds = group.getCandidateAnswer(
                        submitterId, qUid) or []
                    #log(str(sCandidateAnswerIds))
                    for answer in question.listFolderContents():
                        aId = answer.getId()
                        row += [
                            [NaN, [0, 1][aId in sCandidateAnswerIds]][
                            aId in sSuggestedAnswerIds]]
            
            retVal.append( row )

        #log(self.unicodeDecode(retVal))

        return retVal
    
    
    def getItemStatisticsTable(self, keepQuestionGroups=False):
        data = self.getItemStatistics2(keepQuestionGroups)
        participants  = self.getParticipantIds()
        submitters    = filter((lambda candidateId:
                                (candidateId in participants)),
                               self.getSubmitterIds())
        submitters.sort()

        #
        questions = self.getAllQuestions()

        if keepQuestionGroups:
            questions += self.getQuestionGroups()
        else:
            for group in self.getQuestionGroups():
                questions += group.getAllQuestions()

        infoByQid = {}
        for q in questions:
            if hasattr(q, 'getPoints'):
                infoByQid[q.getId()] = {'max': q.getPoints(),
                                        'title': q.getTitle()}
            else:
                infoByQid[q.getId()] = {'max': 'unknown',
                                        'title': q.getTitle()}
        #

        table = []
        #table.append(['questions'] + data.keys())

        row = ['_items']
        for qid in data.keys():
            row.append(infoByQid[qid]['title'])
        table.append(row)

        row = ['_ids']
        for qid in data.keys():
            row.append(qid)
        table.append(row)

        row = ['_max']
        for qid in data.keys():
            row.append(infoByQid[qid]['max'])
        table.append(row)

        for candidateId in submitters:
            row = [candidateId] + ['-' for i in range(len(data.keys()))]
            
            for i in range(len(data.keys())):
                for submission in data[data.keys()[i]]:
                    if submission[0] == candidateId:
                        row[i + 1] = submission[1]
            
            table.append(row)
        
        return table
    

    def getDetailedScores(self):
        llsmc_tool = getToolByName(self, 'llsmc_tool')
        
        submitters = self.getSubmitterIds()
        submitters.sort(lambda a, b: cmp(llsmc_tool.getFullNameById(a),
                                         llsmc_tool.getFullNameById(b)))
        
        table = []

        labels = {}
        for msgid in [('candidate', 'Candidate'),
                      ('test_results_max_score', 'Max. Score'),
                      ('N', 'Number'),
                      ('sample', 'Sample'),
                      ('mean', 'Mean'),
                      ('median', 'Median'),
                      ('stddev', 'Std. Dev.')]:
            labels[msgid[0]] = self.translate(msgid = msgid[0],
                                              domain = I18N_DOMAIN,
                                              default = msgid[1])

        header    = [(labels['candidate'], '', '')]
        maxScores = [labels['test_results_max_score']]

        for group in [self] + self.getQuestionGroups():
            header.append((group.title_or_id(), group.portal_type,
                           group.reference_url()))
            if submitters:
                maxScores.append(group.getPossiblePoints(submitters[0]))
            else:
                maxScores.append('-')
            
            allQuestions = group.getAllQuestions()
            numQuestions = len(allQuestions)
            
            for question in allQuestions:
                header.append((question.title_or_id(), question.portal_type,
                               question.reference_url()))
                maxScores.append(question.getPoints())
        
        table.append(header)
        
        if not submitters:
            return table
        
        for cand in submitters:
            candRow = [cand]
            
            for group in [self] + self.getQuestionGroups():
                candRow.append(group.getCandidatePoints(cand))
                
                allQuestions = group.getAllQuestions()
                numQuestions = len(allQuestions)
                
                for question in allQuestions:
                    candRow.append(question.getCandidatePoints(cand))
                
            table.append(candRow)
        
        # Prepare statistics
        
        temp = ['']
        stats_input = []
        for col in range(1, len(header)):
            for row in table[1:]:
                stats_input.append(row[col])
            temp.append(stats_input)
            stats_input = []
        
        table.append(maxScores)
        
        for analysis in ['mean', 'median', 'stddev']:
            statsRow = [labels[analysis]]
            for set in temp[1:]:
                stats = Statistics(set)
                statsRow.append(getattr(stats, analysis))
            table.append(statsRow)

        return table
    

    def getItemStatistics3(self):
        participants = self.getParticipantIds()
        table = []

        for group in [self] + self.getQuestionGroups():
            table.append(group.title_or_id)
            allQuestions = group.getAllQuestions()
            numQuestions = len(allQuestions)
            
            for question in allQuestions:
                numAnswers = len(question.listFolderContents())
                header = [question.title_or_id(), 'N', '-'] + \
                         [str(i + 1) for i in range(numAnswers)] + \
                         ['mean', 'stddev', 'median', 'mode']

                i = 3
                for answer in question.listFolderContents():
                    if answer.isCorrect():
                        header[i] += '*'
                    i += 1
                
                
                qid = question.UID()
                submitters = filter((lambda candidateId:
                                     (candidateId in participants)),
                                    self.getSubmitterIds())
                afreqs = [question.title_or_id(), 0, 0] + \
                         [0 for i in range(numAnswers)]
                
                for submitter in submitters:
                    sSuggestedAnswerIds = group.getAnswers(submitter, qid) \
                                          or []
                    sCandidateAnswerIds = group.getCandidateAnswer(submitter,
                                                                   qid) or []
                    
                    if not sCandidateAnswerIds:
                        afreqs[2] += 1
                    else:
                        afreqs[1] += 1 # Number of candidates

                        i = 3
                        for answer in question.listFolderContents():
                            aid = answer.getId()
                            afreqs[i] += [0, 1][aid in sCandidateAnswerIds]
                            i += 1

                # Prepare statistics
                i = 1
                items = []
                for freq in afreqs[3:]:
                    items.extend([i for n in range(0, freq)])
                    i += 1

                stats = Statistics(items)
                afreqs.append(stats.mean)
                afreqs.append(stats.stddev)
                afreqs.append(stats.median)
                afreqs.append(stats.mode)

                table.append(header)
                table.append(afreqs)
                    
        return table

    def getItemStatistics2(self, keepQuestionGroups=False):
        participants  = self.getParticipantIds()
        submitters    = filter((lambda candidateId:
                                (candidateId in participants)),
                               self.getSubmitterIds())
        submitters.sort()

        questions = self.getAllQuestions()

        if keepQuestionGroups:
            questions += self.getQuestionGroups()
        else:
            for group in self.getQuestionGroups():
                questions += group.getAllQuestions()

        itemStats = {}
        for q in questions:
            itemStats[q.getId()] = []

        for candidateId in submitters:
            candQuestions = self.getQuestions(candidateId)

            for group in self.getQuestionGroups():
                if keepQuestionGroups:
                    candQuestions += self.getQuestionGroups()
                else:
                    candQuestions += group.getQuestions(candidateId)
            
            for q in candQuestions:
                if hasattr(q, 'getPoints'):
                    max = q.getPoints()                    # Question
                else:
                    max = q.getPossiblePoints(candidateId) # Question group
                
                itemStats[q.getId()].append((candidateId,
                                             q.getCandidatePoints(candidateId),
                                             max))
        
        return itemStats
    
    def getItemStatistics(self, keepQuestionGroups=False):
        participants  = self.getParticipantIds()
        submitters    = filter((lambda candidateId:
                                (candidateId in participants)),
                               self.getSubmitterIds())
        submitters.sort()

        x = {}
        for candidateId in submitters:
            x[candidateId] = []
            
            questions      = self.getQuestions(candidateId)
            questionGroups = self.getQuestionGroups()
            for q in questions:
                max = q.getPoints()
                x[candidateId].append((q.getTitle(),
                                       q.getCandidatePoints(candidateId),
                                       max))
            for q in questionGroups:
                if keepQuestionGroups:
                    max = q.getPossiblePoints(candidateId)
                    x[candidateId].append((q.getTitle(),
                                           q.getCandidatePoints(candidateId),
                                           max))
                else:
                    contained = q.getQuestions(candidateId)
                    for q in contained:
                        max = q.getPoints()
                        x[candidateId].append((q.getTitle(),
                                               q.getCandidatePoints(candidateId),
                                               max))
        return x
    
            
    def getQuestionGroups(self):
        """ Returns all the QuestionGroup elements (no single questions). """
        return filter(lambda item :
                      item.Type() == QuestionGroup.archetype_name,
                      self.listFolderContents())


    security.declarePublic('isEmpty')
    def isEmpty(self):
        return (not self.getQuestionGroups()) and \
               (not self.getAllQuestions())
        
        
    def getPossiblePoints(self, candidateId):
        """ Return how many points the candidate could have got.
            
            @param candidateId the user ID of the candidate whose points you 
                    want to know. 
        """
        points = AbstractQuestionContainer.getPossiblePoints(self, candidateId)
        if(not isNumeric(points)):
            return None
        questionGroups = self.getQuestionGroups()
        for questionGroup in questionGroups:
            questionGroupPoints = questionGroup.getPossiblePoints(candidateId)
            if(not isNumeric(questionGroupPoints)):
                return None
            points = points + questionGroupPoints
        return points
            
            
    def getCandidatePoints(self, candidateId):
        """ Return how many points the candidate got for this test.
            
            @param candidateId the user ID of the candidate whose
            points you want to know.
            
            If a custom evaluation script has been uploaded it will be
            invoked. Otherwise a default method will be used. 
        """
        customScript = self.getEvaluationScript(self.archetype_name)
        questions = self.getQuestions(candidateId)
        questionGroups = self.getQuestionGroups()
        if not customScript: # default
            points = 0
            for questionGroup in questionGroups:
                questionGroupCandidatePoints = \
                    questionGroup.getCandidatePoints(candidateId)
                if not isNumeric(questionGroupCandidatePoints):
                    return None
                points = points + questionGroupCandidatePoints
            for question in questions:
                #if not ('getCandidatePoints' in dir(question)):
                #    return None
                try:
                    question.aq_acquire('getCandidatePoints')
                except:
                    return None
                questionCandidatePoints = \
                    question.getCandidatePoints(candidateId)
                if not isNumeric(questionCandidatePoints):
                    return None
                points = points + questionCandidatePoints
            return points
        else: # custom
            return evalFunString(
                customScript, CUSTOM_EVALUATION_FUNCTION_NAME,
                [self, candidateId, questionGroups, questions])
            
                
    def getWorkflowState(self):
        """ Determine the Plone workflow state. """
        workflow_tool = getToolByName(self, 'portal_workflow')
        return workflow_tool.getInfoFor(self, 'review_state', None)
        
        
    # This function has to be implemented by classes derived from 
    # AbstractQuestionContainer
    security.declarePublic('isPublic')
    def isPublic(self):
        """ Determine whether this test has been published. """
        #~ log('MultipleChoiceTest.isPublic():\n')
        try:
            user = getSecurityManager().getUser()
            showUnpublishedContent = user.has_role('Authenticated') and \
                                     ('Manager' in self.get_local_roles_for_userid(user.getId()) \
                                      or user.has_role('Manager'))

            # log('\tshowUnpublishedContent = %s\n'
            #     %(str(showUnpublishedContent)))
            if showUnpublishedContent:
                return True
            if self.getWorkflowState() != 'published':
                #~ log("\t!'published'\n")
                return False
            now            = DateTime()
            startPublished = getattr(self, 'effective_date', None)
            endPublished   = getattr(self, 'expiration_date', None)
            #~ log('\tnow = ' + str(now) + '\n')
            #~ log('\tstartPublished = ' + str(startPublished) + '\n')
            #~ log('\tendPublished = ' + str(endPublished) + '\n')
            if((startPublished != None) and (startPublished > now)):
                #~ log("\tstartPublished > now\n")
                return False
            if((endPublished != None) and (now > endPublished)):
                #~ log("\tnow > endPublished\n")
                return False
            #~ log("\treturn True\n")
            return True
        except:
            # log("MultipleChoiceTest.isPublic() failed. An unknown "
            #     "exception occurred.\n")
            # log("\treturn False\n")
            return False        
        
        
    def deleteResultOf(self, candidateId):
        """ Deletes any saved information for candidate "candidateId", i.e.
            the test generated for him/her and his/her test results.
        """
        self.setResult(candidateId, None)
            
        
    def pre_validate(self, REQUEST, errors):
        """ This function is called when submitting the edit form (base_edit).
            
            @param REQUEST Contains the input for the form.
            @param errors If any validation error occurs, put it here.
                
            Requests from the 'base_edit' form to delete custom 
            evaluation scripts (from ObjectField 'evaluationScripts') 
            are processed here. 
        """
        if(REQUEST.get('deleteCustomEvaluationScripts', []) != []):
            scriptsToBeDeleted = REQUEST.get('deleteCustomEvaluationScripts',
                                             [])
            if( type(scriptsToBeDeleted) != list ):
                scriptsToBeDeleted = [scriptsToBeDeleted]
            evaluationScripts = self.getEvaluationScripts()
            for archetype_name in scriptsToBeDeleted:
                if(evaluationScripts.has_key(archetype_name)):
                    evaluationScripts.pop(archetype_name)
            self.setEvaluationScripts(evaluationScripts)
            
            
    #~ def post_validate(self, REQUEST, errors):
        #~ if(REQUEST.get('sourceFile_file', False)):
            #~ sourceFile = REQUEST.get('sourceFile_file')
            #~ try:
                #~ string = sourceFile.filename.read()
            #~ except:
                #~ string = sourceFile.read()
            #~ errorMsg = self.processQTIImport(string)
            #~ if(errorMsg):
                #~ errors['sourceFile'] = errorMsg + ' File=' + string
            #~ self.setSourceFile(None)
            
            
# Register this type in Zope
registerTypeLogged(MultipleChoiceTest)
