# -*- coding: iso-8859-1 -*-
#
# $Id: BaseQuestion.py,v 1.13 2006/01/31 18:17:25 wfenske Exp $
#
# Copyright  2004 Otto-von-Guericke-Universitt Magdeburg
#
# This file is part of LlsMultipleChoice.
#
# LlsMultipleChoice is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LlsMultipleChoice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LlsMultipleChoice; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import random

from AccessControl import ClassSecurityInfo

from Acquisition import aq_base, aq_acquire, aq_inner, aq_parent

from Products.Archetypes.public import OrderedBaseFolder, BaseFolderSchema, \
TextField, BaseContent, BaseSchema, Schema, BooleanField, BooleanWidget, IntegerField, IntegerWidget, StringField, \
SelectionWidget, TextAreaWidget, StringWidget, RichWidget
#~ from Products.Archetypes.Marshall import PrimaryFieldMarshaller

from Products.LlsMultipleChoice.config import PROJECTNAME, SKINS_DIR, \
    QUESTION_DIR, QUESTION_TYPES, ANSWER_DIR, \
    ANSWER_TYPES, GLOBALS, PERMISSION_INTERROGATOR, \
    PERMISSION_STUDENT, CUSTOM_EVALUATION_FUNCTION_NAME, I18N_DOMAIN
from Products.LlsMultipleChoice.tools import getParent, filterById, \
     evalFunString, log, hideIdField, registerTypeLogged
from Products.LlsMultipleChoice.AnswerTypes.BaseAnswer import BaseAnswer
from Products.LlsMultipleChoice.InlineTextField import InlineTextField


class BaseQuestion(OrderedBaseFolder):
    """ A very basic question type without any evaluation functions.
          A question is basically a question text and a set of possible
        answers to this question or, in Plone terms, a folder with some
        additional properties such as the question text.
    """

    # See comments in 'AbstractQuestionContainer' for an explanation
    # of the function of a schema, therein defined properties (fields) and
    # internationalization of the widgets
    schema = OrderedBaseFolder.schema + Schema((
            StringField('title',
                required=False,
                searchable=True,
                default='Question',
                widget=StringWidget(
                    label_msgid='label_title',
                    description_msgid='title',
                    i18n_domain='plone'),
            ),
            InlineTextField('question', # See 'description' property of the widget.
                searchable=True,
                required=True,
                primary=True,
                allowable_content_types=('text/plain',
                    'text/structured',
                    'text/restructured',
                    'text/html',),
                default_output_type='text/html',
                widget=TextAreaWidget(
                    label='Question',
                    label_msgid='question_label',
                    description='The question text. This is what the candidate will see.',
                    description_msgid='question_tool_tip',
                    i18n_domain=I18N_DOMAIN),
                validators=('isXML',),
            ),
            BooleanField("randomOrder", # See 'description' property of the widget.
                accessor='isRandomOrder',
                required=False,
                default=1,
                read_permission=PERMISSION_INTERROGATOR,
                widget=BooleanWidget(
                    label='Randomize Answer Order',
                    label_msgid='randomize_answer_order_label',
                    description='Check this box if you want the answers to this question '\
                        'to appear in a different, random order for each candidate. Otherwise the '\
                        'same order as in the &quot;contents&quot;-view will be used.',
                    description_msgid='randomize_answer_order_tool_tip',
                    i18n_domain=I18N_DOMAIN),
            ),
            IntegerField("numberOfRandomAnswers", # See 'description' property of the widget.
                default=-1,
                read_permission=PERMISSION_INTERROGATOR,
                widget=IntegerWidget(
                    label='Number of Random Answers',
                    label_msgid='number_of_random_answers_label',
                    description='The number of answers which are randomly selected when a new test is '\
                        'generated for a candidate. (This only works if &quot;Randomize Answer Order&quot; '\
                        'is checked.)  A value &lt;= 0 means that all answers will be used.',
                    description_msgid='number_of_random_answers_tool_tip',
                    i18n_domain=I18N_DOMAIN),
            ),
        ),
        #~ marshall=PrimaryFieldMarshaller(),
    )
    #~ schema = hideIdField(schema)

    # Use a custom page template for viewing.
    actions = (
        {
            'id': 'view',
            'action': 'string:${object_url}/basequestion_view',
            'permissions': (PERMISSION_INTERROGATOR,),
        },
    )

    # This prevents the Questions from showing up as a portal content type
    global_allow = False
    # Only Answer objects may be put into this folder.
    allowed_content_types = (BaseAnswer.archetype_name,)
    meta_type = portal_type = archetype_name = 'Base Question'
    content_icon = 'Question.png'

    security = ClassSecurityInfo()

    security.declareProtected(PERMISSION_INTERROGATOR, 'contentValues')
    security.declareProtected(PERMISSION_INTERROGATOR, 'listFolderContents')
    # Declaring "folderlistingFolderContents" as protected prevents
    # the answers from being listed if someone without
    # PERMISSION_INTERROGATOR tries to call the "base_view" template
    # for a question.
    security.declareProtected(PERMISSION_INTERROGATOR,
                              'folderlistingFolderContents')


    def getCandidateAnswer(self, candidateId):
        """ Return the candiate's answer to this question.

            @param candidateId The user ID of the candidate.
        """
        parent = getParent(self)
        return parent.getCandidateAnswer(candidateId, self.UID())


    def setCandidateAnswer(self, candidateId, answerValue):
        """ Save the candiate's answer to this question.

            @param candidateId The user ID of the candidate.
            @param answerValue The ID or list of IDs of the seletced answer(s).
        """
        parent = getParent(self)
        parent.setCandidateAnswer(candidateId, self.UID(), answerValue)


    def getAnswers(self, candidateId):
        """ Return a list with the actual answer objects that were
            presented to the candidate with ID candidateId.

            @param candidateId The user ID of the candidate.
        """
        parent = getParent(self)
        allAnswers = self.listFolderContents()
        if(not parent.isPublic()):
            suggestedAnswers = allAnswers
        else:
            # If the candidate has not already seen this test,
            # randomly generate a new one else present the random test
            # that was first generated for him/her
            oldSuggestedAnswerIds = parent.getAnswers(candidateId, self.UID())
            if(not oldSuggestedAnswerIds):
                # generate a new test
                if(self.isRandomOrder()):
                    # use random order
                    if(self.getNumberOfRandomAnswers() > 0):
                        # Return only 'numberOfRandomAnswers' many answers
                        # Separate the correct answers
                        correctAnswers = filter((lambda answer :
                                                 answer.isCorrect()) ,
                                                allAnswers)
                        # Determine the maximum number of correct
                        # answers to be listed
                        maxCorrectAnswers = min(
                            len(correctAnswers),
                            self.getNumberOfRandomAnswers())
                        numCorrectAnswers = random.randint(
                            ([1, 1][self.isAllowMultipleSelection()]),
                            maxCorrectAnswers)
                        numOtherAnswers = \
                                        min(len(allAnswers),
                                            self.getNumberOfRandomAnswers()) \
                                            - numCorrectAnswers
                        # Get the randomized correct & other answers
                        randomCorrectAnswers = random.sample(
                            correctAnswers, numCorrectAnswers)
                        otherAnswers = filter((lambda answer :
                                               answer not in
                                               randomCorrectAnswers),
                                              allAnswers)
                        randomOtherAnswers = random.sample(otherAnswers,
                                                           numOtherAnswers)

                        suggestedAnswers = randomCorrectAnswers \
                                           + randomOtherAnswers
                    else:
                        suggestedAnswers = allAnswers
                    # Randomize the answers before returning them
                    suggestedAnswers = random.sample(suggestedAnswers,
                                                     len(suggestedAnswers))
                else:
                    # use the order in the "contents"-view
                    suggestedAnswers = allAnswers
                suggestedAnswerIds = map((lambda answer : answer.getId()),
                                         suggestedAnswers)
                parent.setAnswers(candidateId, self.UID(), suggestedAnswerIds)
            else:
                # reconstruct the old test
                suggestedAnswers = filterById(oldSuggestedAnswerIds,
                                              allAnswers)
        return suggestedAnswers


# Register this type in Zopes
registerTypeLogged(BaseQuestion)
