# -*- coding: iso-8859-1 -*-
#
# $Id: MultipleChoiceQuestion.py,v 1.11 2006/02/28 16:30:26 mxp Exp $
#
# Copyright  2004 Otto-von-Guericke-Universitt Magdeburg
#
# This file is part of LlsMultipleChoice.
#
# LlsMultipleChoice is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LlsMultipleChoice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LlsMultipleChoice; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import random

from AccessControl import ClassSecurityInfo

from Acquisition import aq_base, aq_acquire, aq_inner, aq_parent

from Products.Archetypes.public import BaseFolder, BaseFolderSchema, \
     BaseContent, BaseSchema, Schema, BooleanField, BooleanWidget, \
     IntegerField, IntegerWidget, StringField, TextField, SelectionWidget, \
     TextAreaWidget, StringWidget, RichWidget

from Products.LlsMultipleChoice.config import PROJECTNAME, SKINS_DIR, \
     QUESTION_DIR, QUESTION_TYPES, ANSWER_DIR, ANSWER_TYPES, GLOBALS, \
     PERMISSION_INTERROGATOR, PERMISSION_STUDENT, \
     CUSTOM_EVALUATION_FUNCTION_NAME, I18N_DOMAIN
from Products.LlsMultipleChoice.tools import getParent, evalFunString, log, \
     registerTypeLogged

from Products.LlsMultipleChoice.QuestionTypes.SelectionQuestion \
     import SelectionQuestion
from Products.LlsMultipleChoice.QuestionTypes.PointsQuestion \
     import PointsQuestion
from Products.LlsMultipleChoice.AnswerTypes.MultipleChoiceAnswer \
     import MultipleChoiceAnswer


class MultipleChoiceQuestion(SelectionQuestion, PointsQuestion):
    """A question in a multiple-choice test."""

    """ This class represents a question in a multiple chopice test.
        The answers can be either right or wrong. The candidate has to select
        which answer(s) he/she thinks is (are) correct.
    """

    # See comments in 'AbstractQuestionContainer' for an explanation
    # of the function of a schema, therein defined properties (fields) and
    # internationalization of the widgets
    schema = SelectionQuestion.schema + PointsQuestion.schema

    # Use a custom page template for viewing.
    actions = (
        {
            'id': 'view',
            'action': 'string:${object_url}/multiplechoicequestion_view',
            'permissions': (PERMISSION_INTERROGATOR,),
        },
    )

    # Only Answer objects may be put into this folder.
    allowed_content_types = (MultipleChoiceAnswer.archetype_name,)
    meta_type = portal_type = archetype_name = 'MC Question'

    security = ClassSecurityInfo()

    security.declarePrivate('getCorrectAnswerIds')
    def getCorrectAnswerIds(self, candidateId):
        """ Return the IDs of the correct answers to this question that
            were presented to the candidate with ID candidateId.

            @param candidateId The user ID of the candidate.
        """
        # all the answers to this question
        correctAnswers = self.listFolderContents()
        # filter out the correct ones
        correctAnswers = filter((lambda answer : answer.isCorrect()),
                                correctAnswers)
        # get the IDs of the answers
        correctAnswerIds = map((lambda answer : answer.getId()),
                               correctAnswers)
        # filter out the ones that were presented to the candidate
        parent = getParent(self)
        suggestedAnswerIds = parent.getAnswers(candidateId, self.UID())
        correctSuggestedAnswerIds = filter((lambda answerId :
                                            answerId in suggestedAnswerIds),
                                           correctAnswerIds)
        return correctSuggestedAnswerIds


    def getCandidatePoints(self, candidateId):
        """Return how many points the user got for this question.

        @param candidateId The user ID of the candidate whose points
        you want to know.

        If a custom scoring script has been uploaded it will be
        invoked. Otherwise a default method will be used.
        """
        parent = getParent(self)
        
        # The IDs of the questions the candidate could have selected
        suggestedAnswerIds = parent.getAnswers(candidateId, self.UID())
        # The IDs of answers the candidate should have selected
        correctAnswerIds = self.getCorrectAnswerIds(candidateId)
        correctAnswerIds.sort()
        # The IDs of the answers the candidate did select
        givenAnswerIds = parent.getCandidateAnswer(candidateId, self.UID())
        
        customScript = parent.getEvaluationScript(self.archetype_name)
        if customScript: # use custom script
            return evalFunString(customScript, CUSTOM_EVALUATION_FUNCTION_NAME,
                                 [self, candidateId, givenAnswerIds])
        else: # default

            # The function selected in the edit tab of the test
            evalFun = parent.getScoringFunction()
            
            if evalFun == 'cruel':
                if givenAnswerIds is None:
                    givenAnswerIds = []
                givenAnswerIds.sort()
                # Give all the points if everything was
                # correct. Otherwise, give no points.
                return [0, self.getPoints()][givenAnswerIds
                                             == correctAnswerIds]
            elif evalFun == 'guessing':
                minPoints = 0
                
                if givenAnswerIds is None:
                    return minPoints
                else:
                    numCorrectAnswers = len(correctAnswerIds)
                    numWrongAnswers   = len(suggestedAnswerIds) \
                                        - numCorrectAnswers
                    
                    points =   self.getPoints() * 1.0
                    plus   = 0
                    if numCorrectAnswers:
                        plus = points / numCorrectAnswers
                    minus  = 0
                    if numWrongAnswers:
                        minus = - points / numWrongAnswers 

                    # Loop over the given answers. Detract points for
                    # the wrong choices, add points for the correct
                    # ones.
                    score = 0.0
                    for givenAnswerId in givenAnswerIds:
                        score += [minus, plus][givenAnswerId
                                               in correctAnswerIds]
                    return max(score, minPoints)
            else:
                raise NotImplementedError(u"Scoring Function '%s'"
                                          % evalFun)

# Register this type in Zope
registerTypeLogged(MultipleChoiceQuestion)
